/* I2S es8311 Example

    This example code is in the Public Domain (or CC0 licensed, at your option.)

    Unless required by applicable law or agreed to in writing, this
    software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
    CONDITIONS OF ANY KIND, either express or implied.
*/
#include <stdio.h>
#include <string.h>
#include <sys/unistd.h>
#include <sys/stat.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "driver/i2s.h"
#include "esp_system.h"
#include "esp_log.h"
#include "esp_vfs_fat.h"
#include "sdmmc_cmd.h"
#include "driver/sdmmc_defs.h"
#include "driver/sdmmc_host.h"
#include "es8311.h"

/* I2C port and GPIOs */
#define I2C_NUM         (0)
#define I2C_SDA_IO      (GPIO_NUM_18)
#define I2C_SCL_IO      (GPIO_NUM_23)

/* I2S port and GPIOs */
#define I2S_NUM         (0)
#define I2S_MCK_IO      (GPIO_NUM_0)
#if !I2S_NUM
#define I2S_BCK_IO      (GPIO_NUM_5)
#define I2S_WS_IO       (GPIO_NUM_25)
#define I2S_DO_IO       (GPIO_NUM_26)
#define I2S_DI_IO       (GPIO_NUM_35)
#else
#define I2S_BCK_IO      (GPIO_NUM_32)
#define I2S_WS_IO       (GPIO_NUM_33)
#define I2S_DO_IO       (-1)
#define I2S_DI_IO       (GPIO_NUM_36)
#endif

#define SD_CARD_INTR_GPIO  (GPIO_NUM_34)
#define SD_CARD_PWR_CTRL   (GPIO_NUM_13)



/* Example configurations */
#define EXAMPLE_RECV_BUF_SIZE   (2048)
#define EXAMPLE_SAMPLE_RATE     (96000)
#define EXAMPLE_MCLK_MULTIPLE   I2S_MCLK_MULTIPLE_256
#define EXAMPLE_VOICE_VOLUME    (60)
#define EXAMPLE_MIC_GAIN        ES8311_MIC_GAIN_18DB
#define EXAMPLE_MSG_QUEUE_NUM   (10)
#define EXAMPLE_SAMPLE_BIT      I2S_BITS_PER_SAMPLE_16BIT

#define FLASH_RECORD_SIZE       (EXAMPLE_SAMPLE_RATE * 40)
#define SD_MOUNT_POINT          "/sdcard"
#define SPI_DMA_CHAN            (2)

typedef enum {
    MSG_OWNER_SD_CARD,
    MSG_OWNER_I2S_RX,
} msg_owner_t;

typedef struct {
    uint32_t    len;
    uint8_t     *buf;
} recv_msg_t;


static const char *TAG = "i2s_record";
static const int WAVE_HEADER_SIZE = 44;

// When testing SD and SPI modes, keep in mind that once the card has been
// initialized in SPI mode, it can not be reinitialized in SD mode without
// toggling power to the card.
sdmmc_host_t host = SDMMC_HOST_DEFAULT();
QueueHandle_t msg_que;
sdmmc_card_t *card;


static void es8311_codec_init(void)
{
    /* Initialize I2C peripheral */
    i2c_config_t es_i2c_cfg = {
        .sda_io_num = I2C_SDA_IO,
        .scl_io_num = I2C_SCL_IO,
        .mode = I2C_MODE_MASTER,
        .sda_pullup_en = GPIO_PULLUP_ENABLE,
        .scl_pullup_en = GPIO_PULLUP_ENABLE,
        .master.clk_speed = 100000,
    };
    i2c_param_config(I2C_NUM, &es_i2c_cfg);
    i2c_driver_install(I2C_NUM, I2C_MODE_MASTER,  0, 0, 0);

    /* Initialize es8311 codec */
    es8311_handle_t es_handle = es8311_create(I2C_NUM, ES8311_ADDRRES_0);
    if (es_handle == NULL) {
        ESP_LOGE(TAG, "es8311 create failed");
        abort();
    }
    es8311_clock_config_t es_clk = {
        .mclk_from_mclk_pin = true,
        .sample_frequency = EXAMPLE_SAMPLE_RATE
    };
    es8311_init(es_handle, &es_clk, ES8311_RESOLUTION_16, ES8311_RESOLUTION_16);
    es8311_sample_frequency_config(es_handle, EXAMPLE_SAMPLE_RATE * EXAMPLE_MCLK_MULTIPLE, EXAMPLE_SAMPLE_RATE);
    es8311_microphone_config(es_handle, false);
    es8311_voice_volume_set(es_handle, EXAMPLE_VOICE_VOLUME, NULL);
    es8311_microphone_gain_set(es_handle, EXAMPLE_MIC_GAIN);
}

static void i2s_driver_init(void)
{
    i2s_config_t i2s_cfg = {
        .mode = I2S_MODE_MASTER | I2S_MODE_RX | I2S_MODE_TX,
        .sample_rate = EXAMPLE_SAMPLE_RATE,
        .bits_per_sample = I2S_BITS_PER_SAMPLE_16BIT,
        .channel_format = I2S_CHANNEL_FMT_RIGHT_LEFT,
        .communication_format = I2S_COMM_FORMAT_STAND_I2S,
        .tx_desc_auto_clear = true,
#if SOC_I2S_SUPPORTS_TDM
        .total_chan = 2,
        .chan_mask = I2S_TDM_ACTIVE_CH0 | I2S_TDM_ACTIVE_CH1,
        .left_align = false,
        .big_edin = false,
        .bit_order_msb = false,
        .skip_msk = false,
#endif
        .dma_buf_count = 8,
        .dma_buf_len = 64,
        .use_apll = false,
        .mclk_multiple = EXAMPLE_MCLK_MULTIPLE,
        .intr_alloc_flags = ESP_INTR_FLAG_LEVEL1,
    };

    i2s_driver_install(I2S_NUM, &i2s_cfg, 0, NULL);
    i2s_pin_config_t i2s_pin_cfg = {
        .mck_io_num = I2S_MCK_IO,
        .bck_io_num = I2S_BCK_IO,
        .ws_io_num = I2S_WS_IO,
        .data_out_num = I2S_DO_IO,
        .data_in_num = I2S_DI_IO
    };
    i2s_set_pin(I2S_NUM, &i2s_pin_cfg);
}

static void mount_sdcard(const char *basePath)
{
    gpio_config_t sdcard_pwr_pin_cfg = {
        .pin_bit_mask = 1UL << SD_CARD_PWR_CTRL,
                            .mode = GPIO_MODE_OUTPUT,
                            .pull_up_en = GPIO_PULLUP_DISABLE,
                            .pull_down_en = GPIO_PULLDOWN_DISABLE,
                            .intr_type = GPIO_INTR_DISABLE,
    };
    gpio_config(&sdcard_pwr_pin_cfg);
    gpio_set_level(SD_CARD_PWR_CTRL, 0);

    host.flags = SDMMC_HOST_FLAG_1BIT;

    sdmmc_slot_config_t slot_config = SDMMC_SLOT_CONFIG_DEFAULT();
    slot_config.gpio_cd = SD_CARD_INTR_GPIO;
    slot_config.width = 1;

    esp_vfs_fat_sdmmc_mount_config_t mount_config = {
        .format_if_mount_failed = true,
        .max_files = 5,
        .allocation_unit_size = 16 * 1024
    };


    esp_err_t ret;
retry:
    ret = esp_vfs_fat_sdmmc_mount(basePath, &host, &slot_config, &mount_config, &card);

    if (ret != ESP_OK) {
        if (ret == ESP_FAIL) {
            ESP_LOGE(TAG, "Failed to mount filesystem. "
                     "If you want the card to be formatted, set the EXAMPLE_FORMAT_IF_MOUNT_FAILED menuconfig option.");
        } else {
            ESP_LOGE(TAG, "Failed to initialize the card (%s). "
                     "Make sure SD card lines have pull-up resistors in place.", esp_err_to_name(ret));
        }
        ESP_LOGI(TAG, "retry in 3 sec");
        vTaskDelay(300);
        goto retry;
    }
    sdmmc_card_print_info(stdout, card);
}

static void sd_record_task(void *args)
{
    // Use POSIX and C standard library functions to work with files.
    int flash_wr_size = 0;

    recv_msg_t msg;
    char wav_header_fmt[WAVE_HEADER_SIZE];
    const char filename[] = SD_MOUNT_POINT"/record";

retry:
    ESP_LOGI(TAG, "Opening new file");
    // First check if file exists before creating a new file.
    struct stat st;
    printf("Filename: %s\n", filename);
    if (stat(filename, &st) == 0) {
        // Delete it if it exists
        unlink(filename);
    }

    // Create new WAV file
    FILE *f = fopen(filename, "w");
    if (f == NULL) {
        ESP_LOGE(TAG, "Failed to open file for writing");
        vTaskDelay(300);
        goto retry;
    }
    // Write the header to the WAV file
    // wavHeader(wav_header_fmt, FLASH_RECORD_SIZE, EXAMPLE_SAMPLE_RATE);
    // fwrite(wav_header_fmt, 1, WAVE_HEADER_SIZE, f);
    while (flash_wr_size < FLASH_RECORD_SIZE) {
        // receive from evnet queue
        xQueueReceive(msg_que, &msg, portMAX_DELAY);
        // Write the samples to the WAV file
        fwrite(msg.buf, 1, msg.len, f);
        flash_wr_size += msg.len;
    }
    ESP_LOGI(TAG, "Recording done!");
    fclose(f);
    esp_vfs_fat_sdcard_unmount(SD_MOUNT_POINT, card);
    ESP_LOGI(TAG, "File written on SDCard");
    while (1) {
        vTaskDelay(500);
    }
}

static void i2s_read_task(void *args)
{
    recv_msg_t msg[EXAMPLE_MSG_QUEUE_NUM];
    for (int i = 0; i < EXAMPLE_MSG_QUEUE_NUM; i++) {
        msg[i].len = 0;
        msg[i].buf = calloc(EXAMPLE_RECV_BUF_SIZE, sizeof(uint8_t));
        if (!msg[i].buf) {
            ESP_LOGE(TAG, "No memory for read data buffer");
            abort();
        }
    }
    uint8_t cnt = 0;
    uint32_t wirte_num;
    while (1) {
        /* Read sample data from mic */
        i2s_read(I2S_NUM, msg[cnt].buf, EXAMPLE_RECV_BUF_SIZE, &msg[cnt].len, portMAX_DELAY);
        xQueueSend(msg_que, &msg[cnt], portMAX_DELAY);
        cnt++;
        cnt %= EXAMPLE_MSG_QUEUE_NUM;
    }
}

void app_main(void)
{
    mount_sdcard(SD_MOUNT_POINT);
    /* Initialize i2s peripheral */
    i2s_driver_init();
    /* Initialize i2c peripheral and config es8311 codec by i2c */
    es8311_codec_init();
    /* message queue */
    msg_que = xQueueCreate(EXAMPLE_MSG_QUEUE_NUM, sizeof(recv_msg_t));
    /* Record sample data in SD card */
    xTaskCreatePinnedToCore(sd_record_task, "sd_record_task", 4096, NULL, 5, NULL, 1);
    /* Read ADC by I2S */
    xTaskCreatePinnedToCore(i2s_read_task, "i2s_read_task", 8192, NULL, 5, NULL, 0);
}
